/*
 * WALTZ
 *
 * In the header file there are definitions of structures, types, macros,
 * global arrays and variables used throughout the library.
 *
 * Copyright  2012-2013 Pavel Vesely <vesely@kam.mff.cuni.cz>
 *
 * This program is distributed under the GNU General Public License.
 * See COPYING for details.
 */

#ifndef TZAARLIB_H_INCLUDED
#define TZAARLIB_H_INCLUDED

#include "build/config.h"
#include <stdint.h>
#include <stdbool.h>
#include "build/types.h"

typedef u64 thash;

//helpful macros
#define MALLOC(a) ((a *)malloc(sizeof(a)))
#define FOR(i,m,n) for (int i=m; i<n; i++)
#define SIGND(a,b) ((a) == (b) ? 0 : ((a) < (b) ? 1 : -1))
#define SIGN(a) ((a) > 0 ? 1 : ((a) < 0 ? -1 : 0))
#define MAX(a,b) ((a) >= (b) ? (a) : (b))
#define MIN(a,b) ((a) <= (b) ? (a) : (b))

#ifdef __GNUC__
#define UNUSED __attribute__((unused))
#define ALWAYS_INLINE inline __attribute__((always_inline))
#else
#define UNUSED
#define ALWAYS_INLINE inline
#endif

#ifdef DEBUG
	// Prints a message to the output. Flushing the buffer is because of program termination.
	#define DPRINT(...) do { fprintf(stderr, "D: "); fprintf(stderr, ## __VA_ARGS__); fprintf(stderr, "\n"); fflush(stderr); } while(0)
	// Checks the condition and if it does not hold, prints a message.
	#define ASSERT(con,...) do { if (!(con)) { fprintf(stderr, "ASSERTION FAILED %s:%d: ", __FILE__, __LINE__); fprintf(stderr, ## __VA_ARGS__); fprintf(stderr, "\n"); fflush(stderr); }} while(0)
	// Command used only in the case of debugging
	#define DBG(cmd) { cmd; }
	// Command used only in the case of debugging, this macro is used for declarations
	#define DBGDECL(cmd) cmd;
#else
	#define DPRINT(...) do { } while(0)
	#define DPRINT2(...) do { } while(0)
	#define ASSERT(...) do { } while(0)
	#define DBG(cmd) do {} while(0)
	#define DBGDECL(cmd) do {} while(0)
#endif

// logging
#ifdef SEARCHFORBUG
// these macros are used only when we want to find a bug
	#define DPRINT2(format,...) do { fprintf(stderr, "D: "); fprintf(stderr, format, ## __VA_ARGS__); fprintf(stderr, "\n"); fflush(stderr); } while(0)
	#define ASSERT2(con,...) do { if (!(con)) { fprintf(stderr, "ASSERTION FAILED %s:%d: ", __FILE__, __LINE__); fprintf(stderr, ## __VA_ARGS__); fprintf(stderr, "\n"); fflush(stderr); }} while(0)
	#define DBG2(cmd) do { cmd; } while(0)
#else
	#define DPRINT2(...) do { } while(0)
	#define ASSERT2(...) do { } while(0)
	#define DBG2(cmd)  do { } while(0)
#endif

// -----------------
// CONSTANTS
// ----------------
#define BOARD_ARRAY_SIZE 81
#define INVALID_INPUT -1

// types of fields
#define EMPTY 0
#define BORDER 100
#define TOTT 1
#define TZARRA 2
#define TZAAR 3

//counts
#define CTOTTS 15
#define CTZARRAS 9
#define CTZAARS 6
#define TOTAL_STONES 60		//for both players
#define STONE_TYPES 7
#define MAX_STACK_HEIGHT 15
#define MAX_MOVES 121		//per moveNumber -- it's 60 * 2 (capture + pass)
#define MAX_POSSIBILITIES 200	// per move of a player, it's 30*6 (max. 30 stones per player and 6 directions)

// players
#define WHITE 1
#define BLACK -1

//standard return values
#define OK 0
#define ERROR 1
#define CANNOT_LOAD_POSITION 2
#define BAD_STONE 3
#define CANNOT_SAVE_POSITION 4
#define GET_BEST_MOVE_ERROR 5
#define FILE_WITH_POSITION_NULL 6
#define FILE_FOR_BEST_MOVES_NULL 7

//AI constants
#define WIN 2000000000

#define ALPHABETA_DEPTH 5	// for simple AB without ID
#define AI_TIME_LIMIT 30
#define AI_RANDOM_MARGIN 20
#define AI_RANDOM_MARGIN_BIGGER 5000
#define MIN_AB_DEPTH 5
#define BEGINNER_AB_DEPTH 4
#define INTERMEDIATE_AB_DEPTH 5

//types of AI
#define HUMAN 0			// not supported :)
#define AIALPHABETA 1
#define AIALPHABETA_ID 2
#define AIALPHABETA_ID_PV 3
#define AIALPHABETA_ID_PV_MO 4
#define AIALPHABETA_ID_MO 5
#define AIALPHABETA_RANDOM 6
#define AIALPHABETA_ID_PV_MO_SCOUT 7
#define AIALPHABETA_ID_PV_MO_HISTORY 8
#define AIALPHABETA_ID_PV_MO_SCOUT_HISTORY 9
#define AIALPHABETA_MAX 9
#define AIALPHABETA_BEST 9
#define DFPNS 20
#define DFPNS_EPS_TRICK 21
#define WEAK_PNS 22
#define DFPNS_EVAL_BASED 23
#define DFPNS_WEAK_EPS_EVAL 24
#define DFPNS_DYNAMIC_WIDENING_EPS_EVAL 25
#define DFPNS_BEST 24
#define DFPNS_MAX 25
#define BEGINNERS_AI 40		// level challenging for beginners
#define INTERMEDIATE_AI 41	// for intermediate players
#define AICOMBI_RANDOM_AB_PNS 42 // for experts, the best
#define MAIN_AI 42

extern const int BranchingFactorByFreeFields[];
extern const int AB_ID_timeMultByFreeFieldsEvenDepth[];
extern const int AB_ID_timeMultByFreeFieldsOddDepth[];
extern const int InitialStoneCounts[];

typedef struct move {
	struct move *next;	//for generating, not history
	s32 from, to;		//pass move <=> from == -1 and to == -1;
	s32 oldStackHeight, oldStone;
	s32 value;		// value of the move; after executing it is used to remember old materialValue
} Move;

typedef struct fullMove {	//for pns
	Move *m1, *m2;
} FullMove;

// Position reprezentation
extern int board[BOARD_ARRAY_SIZE];
extern int stackHeights[BOARD_ARRAY_SIZE];
extern int player, moveNumber;

// History
extern int turnNumber;
extern Move *history[MAX_MOVES];

// Position properties useful for evaluation function and moves sorting
extern int counts[STONE_TYPES];
extern int value, materialValue, stoneSum;
extern thash hash;
extern int highestStack[STONE_TYPES];
extern int highestStackPos[STONE_TYPES];
extern int highestStackPosCache[STONE_TYPES][MAX_STACK_HEIGHT];
extern int stackPosByHeight[STONE_TYPES][MAX_STACK_HEIGHT];
extern int countsByHeight[STONE_TYPES][MAX_STACK_HEIGHT];
extern int zoneOfControl[STONE_TYPES], threatenByCounts[BOARD_ARRAY_SIZE];

// For saving
extern double searchDuration;

// For searching (AB and PNS)
extern int currDepth;		//for tests on AB
extern uns searchedNodes;

// Debug constants
#ifdef DEBUG
extern int moveAlive, entryAlive, ttHit, ttFound, ttKick, prunedCount, entry2Alive, tt2Kick, tt2Hit, tt2Found, ttCollision;
#endif

// ---------------
// FUNCTIONS -- TZAAR LIB
// ---------------
int GetBestMove(int ai, int time, int depthAB, bool recallAI, Move **move1, Move **move2);

#include "ai/moves.h"
#include "ai/init.h"
#include "ai/pns.h"
#include "ai/alphabeta.h"

#endif				// TZAARLIB_H_INCLUDED
