/*
 * WALTZ
 *
 * The module moves contains functions for generating, executing and
 * reverting moves. There are also helping functions for deallocating memory
 * (free a~single move or a~linked list of moves), determining whether someone
 * won in the current position, updating the zone of control after executing
 * or reverting move and converting between a~field index and a~field name
 * (for example field on index 3 has name D1).
 *
 * Copyright  2012-2013 Pavel Vesely <vesely@kam.mff.cuni.cz>
 *
 * This program is distributed under the GNU General Public License.
 * See COPYING for details.
 */

#include "ai/moves.h"

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>


const int CapturingStackHeightAdvantage[] = { 0, 0, 15, 50, 160, 200, 210, 220, 230, 240, 250, 260, 270, 280, 290, 300, 310 };

// difference in x direction
const int dxs[] = { 1, -1, 1, -1, 0, 0 };
// difference in y direction
const int dys[] = { 1, -1, 0, 0, 1, -1 };

// for converting between field index and name
static char *FieldNames[] = {
	"A1", "B1", "C1", "D1", "E1", "-" , "-" , "-" , "-" ,
	"A2", "B2", "C2", "D2", "E2", "F1", "-" , "-" , "-" ,
	"A3", "B3", "C3", "D3", "E3", "F2", "G1", "-" , "-" ,
	"A4", "B4", "C4", "D4", "E4", "F3", "G2", "H1", "-" ,
	"A5", "B5", "C5", "D5", "-" , "F4", "G3", "H2", "I1",
	"-" , "B6", "C6", "D6", "E5", "F5", "G4", "H3", "I2",
	"-" , "-" , "C7", "D7", "E6", "F6", "G5", "H4", "I3",
	"-" , "-" , "-" , "D8", "E7", "F7", "G6", "H5", "I4",
	"-" , "-" , "-" , "-" , "E8", "F8", "G7", "H6", "I5"
};

/// Returns the value of a stack on the field fieldTo with the height stackHeight
ALWAYS_INLINE
int StackQuality(int stackHeight, int stone, int fieldTo)
{
	return StackHeightValue[stackHeight] * StackByCountValue[counts[stone + 3]] +
	    StackValueByField[stackHeight][fieldTo] * FIELD_VALUE_MULT;
}

/// Updating ZOC after removing a stone on the field to
ALWAYS_INLINE
void ZOCRemovingStoneTo(int field)
{				// supposing that stone is still on its old place ...
	FOR(j, 0, DIRECTION_COUNT) {
		int cx = field % 9 + dxs[j];
		int cy = field / 9 + dys[j];
		int curr = cy * 9 + cx;
		while (curr >= 0 && curr < BOARD_ARRAY_SIZE && board[curr] == EMPTY) {
			cx += dxs[j];
			cy += dys[j];
			curr = cy * 9 + cx;
		}
		if (curr >= 0 && curr < BOARD_ARRAY_SIZE && board[curr] != BORDER && board[field] * board[curr] < 0) {
			if (stackHeights[field] >= stackHeights[curr]) {
				threatenByCounts[curr]--;
				if (threatenByCounts[curr] == 0) {
					zoneOfControl[board[curr] + 3]--;
				}
			}
			if (stackHeights[field] <= stackHeights[curr]) {
				threatenByCounts[field]--;
				if (threatenByCounts[field] == 0) {
					zoneOfControl[board[field] + 3]--;
				}
			}
		}
	}
}

/// Updating ZOC after removing a stone on the field from
ALWAYS_INLINE
void ZOCRemovingStoneFrom(int field, int to)
{				// supposing that stone is still on its old place ...
	int opposites[DIRECTION_COUNT / 2];
	FOR(j, 0, DIRECTION_COUNT / 2) opposites[j] = -1;
	FOR(j, 0, DIRECTION_COUNT) {
		int cx = field % 9 + dxs[j];
		int cy = field / 9 + dys[j];
		int curr = cy * 9 + cx;
		while (curr >= 0 && curr < BOARD_ARRAY_SIZE && board[curr] == EMPTY) {
			cx += dxs[j];
			cy += dys[j];
			curr = cy * 9 + cx;
		}
		if (curr >= 0 && curr < BOARD_ARRAY_SIZE && board[curr] != BORDER) {
			int o = opposites[j / 2];
			if (o == -1)
				opposites[j / 2] = curr;
			else if (board[o] * board[curr] < 0 && o != to && curr != to) {	// stones hidden by removed stone
				if (stackHeights[o] >= stackHeights[curr]) {
					if (threatenByCounts[curr] == 0) {
						zoneOfControl[board[curr] + 3]++;
					}
					threatenByCounts[curr]++;
				}
				if (stackHeights[o] <= stackHeights[curr]) {
					if (threatenByCounts[o] == 0) {
						zoneOfControl[board[o] + 3]++;
					}
					threatenByCounts[o]++;
				}
			}
			if (board[field] * board[curr] < 0) {
				if (stackHeights[field] >= stackHeights[curr]) {
					threatenByCounts[curr]--;
					if (threatenByCounts[curr] == 0) {
						zoneOfControl[board[curr] + 3]--;
					}
				}
				if (stackHeights[field] <= stackHeights[curr]) {
					threatenByCounts[field]--;
					if (threatenByCounts[field] == 0) {
						zoneOfControl[board[field] + 3]--;
					}
				}
			}
		}
	}
}

/// Updating ZOC after adding a stone on the field from
ALWAYS_INLINE
void ZOCAddedStoneFrom(int field, int to)
{				// supposing stone is already on its place
	int opposites[DIRECTION_COUNT / 2];
	FOR(j, 0, DIRECTION_COUNT / 2) opposites[j] = -1;
	FOR(j, 0, DIRECTION_COUNT) {
		int cx = field % 9 + dxs[j];
		int cy = field / 9 + dys[j];
		int curr = cy * 9 + cx;
		while (curr >= 0 && curr < BOARD_ARRAY_SIZE && board[curr] == EMPTY) {
			cx += dxs[j];
			cy += dys[j];
			curr = cy * 9 + cx;
		}
		if (curr >= 0 && curr < BOARD_ARRAY_SIZE && board[curr] != BORDER) {
			int o = opposites[j / 2];
			if (o == -1)
				opposites[j / 2] = curr;
			else if (board[o] * board[curr] < 0 && o != to && curr != to) {	// stones hidden by added stone
				if (stackHeights[o] >= stackHeights[curr]) {
					threatenByCounts[curr]--;
					if (threatenByCounts[curr] == 0) {
						zoneOfControl[board[curr] + 3]--;
					}
				}
				if (stackHeights[o] <= stackHeights[curr]) {
					threatenByCounts[o]--;
					if (threatenByCounts[o] == 0) {
						zoneOfControl[board[o] + 3]--;
					}
				}
			}
			if (board[field] * board[curr] < 0) {
				if (stackHeights[field] >= stackHeights[curr]) {
					if (threatenByCounts[curr] == 0) {
						zoneOfControl[board[curr] + 3]++;
					}
					threatenByCounts[curr]++;
				}
				if (stackHeights[field] <= stackHeights[curr]) {
					if (threatenByCounts[field] == 0) {
						zoneOfControl[board[field] + 3]++;
					}
					threatenByCounts[field]++;
				}
			}
		}
	}
}

/// Updating ZOC after adding a stone on field to
ALWAYS_INLINE
void ZOCAddedStoneTo(int field)
{				// supposing stone is already on its place
	FOR(j, 0, DIRECTION_COUNT) {
		int cx = field % 9 + dxs[j];
		int cy = field / 9 + dys[j];
		int curr = cy * 9 + cx;
		while (curr >= 0 && curr < BOARD_ARRAY_SIZE && board[curr] == EMPTY) {
			cx += dxs[j];
			cy += dys[j];
			curr = cy * 9 + cx;
		}
		if (curr >= 0 && curr < BOARD_ARRAY_SIZE && board[curr] != BORDER && board[field] * board[curr] < 0) {
			if (stackHeights[field] >= stackHeights[curr]) {
				if (threatenByCounts[curr] == 0) {
					zoneOfControl[board[curr] + 3]++;
				}
				threatenByCounts[curr]++;
			}
			if (stackHeights[field] <= stackHeights[curr]) {
				if (threatenByCounts[field] == 0) {
					zoneOfControl[board[field] + 3]++;
				}
				threatenByCounts[field]++;
			}
		}
	}
}

/// Free a move and update counters
ALWAYS_INLINE
void FreeMove(Move * move)
{
	ASSERT2(move != NULL, "move to free NULL");
	free(move);
	DBG(moveAlive--);
}

/// Free linked list of moves, but only the ones that are different from exception
/// Note: exception == NULL may occure
ALWAYS_INLINE
void FreeAllMoves(Move * move, Move * exception)
{
	Move *tmp;
	while (move != NULL) {
		tmp = move;
		move = move->next;
		if (tmp != exception)
			FreeMove(tmp);
	}
}

/// Free linked list of moves
ALWAYS_INLINE
void FreeAllMovesWithoutException(Move * move)
{
	Move *tmp;
	while (move != NULL) {
		tmp = move;
		move = move->next;
		FreeMove(tmp);
	}
}

/// Returns whether a given move is possible in the current position
ALWAYS_INLINE
int IsMovePossible2(Move * move, int moveNum)
{				//used by AB to determine TT2 collision
	ASSERT2(move != NULL, "IsMovePossible: move NULL");
	int from = move->from, to = move->to;
	if (from == -1) {	//pass
		if (moveNum == 2)
			return true;
		else {
			DPRINT("move not possible: pass in the 1st move"); return false;
		}
	}
	// these exceptional cases shouldn't happen
	ASSERT2(!(from < 0 || from >= BOARD_ARRAY_SIZE || to < 0 || to >= BOARD_ARRAY_SIZE || board[from] == BORDER
		 || board[to] == BORDER), "IsMovePossible: bad move THIS SHOULD NOT HAPPEN; from %d to %d", from, to);
	if (SIGN(board[from]) != player) {
		DPRINT("move not possible: bad color on from field"); return false;
	}
	if (board[to] == EMPTY) {
		DPRINT("move not possible: empty target"); return false; // empty target
	}
	if (board[from] == EMPTY) {
		DPRINT("move not possible: empty source"); return false; // empty source
	}
	int sx = from % 9;
	int sy = from / 9;	// source in the x, y coordinates
	int tx = to % 9;
	int ty = to / 9;	// target in the x, y coordinates

	// source and target aren't on the same line
	if (sx != tx && sy != ty && abs(sx - tx) != abs(sy - ty)) {
		DPRINT("move not possible: source and target aren't on the same line"); return false;
	}
	int dx = SIGND(sx, tx), dy = SIGND(sy, ty);
	if (dx * dy < 0) {
		DPRINT("move not possible: source and target are on bad diagonals"); return false;
	}

	// ensure that all lines between source and target are empty
	int cx = sx, cy = sy;
	while (true) {
		cx += dx;
		cy += dy;
		if (cx == tx && cy == ty)
			break;
		int c = cy * 9 + cx;
		if (c < 0 || c >= BOARD_ARRAY_SIZE) {
			DPRINT("move not possible: out of range while scaning empty places between from and to"); return false;
		}
		if (board[c] != EMPTY) {
			DPRINT("move not possible: a stack between from and to"); return false;
		}

	}
	if (board[from] * board[to] > 0) {	// stacking
		if (counts[board[to] + 3] == 1) {
			DPRINT("move not possible: stacking on the last piece"); return false; // don't stack on the last piece
		}
		if (moveNum == 2)
			return true;	//stacking
		else {
			DPRINT("move not possible: stacking not possible in the 1st move (moveNum %d)", moveNum); return false;
		}
	} else if (stackHeights[from] >= stackHeights[to])
		return true;	// capturing higher stack is not possible
	DPRINT("move not possible: no case of "); return false;
}

/// Returns whether a given move is possible in the current position
ALWAYS_INLINE
int IsMovePossible(Move * move)
{
	return IsMovePossible2(move, moveNumber);
}

void printZOCDebug(void)
{		// used for debugging Zone of Control
	FOR(i, 0, BOARD_ARRAY_SIZE) {
		printf("%d ", threatenByCounts[i]);
		if ((i + 1) % 9 == 0)
			printf("\n");
	}
	printf("\n");
	FOR(i, 0, STONE_TYPES) {
		printf("%d ", zoneOfControl[i]);
	}
	printf("\n");
}

/// When moveNumber is 1, player has to have a possibility to capture, thus this method searches for a possible capture.
/// Note: using ZOC is much more quicker
bool HasLegalMoves(void)
{
	ASSERT2(moveNumber == 1, "hasLegalMoves is useless when moveNumber == 2");
	ASSERT(false, "calling has legal moves");
	for (int i = 0; i < BOARD_ARRAY_SIZE; i++) {
		if (board[i] == BORDER || board[i] == EMPTY || board[i] * player < 0)
			continue;
		for (int j = 0; j < DIRECTION_COUNT; j++) {
			int cx = i % 9 + dxs[j];
			int cy = i / 9 + dys[j];
			int curr = cy * 9 + cx;
			while (curr >= 0 && curr < BOARD_ARRAY_SIZE && board[curr] == EMPTY) {
				cx += dxs[j];
				cy += dys[j];
				curr = cy * 9 + cx;
			}
			if (curr < 0 || curr >= BOARD_ARRAY_SIZE || board[curr] == BORDER)
				continue;
			if (board[i] * board[curr] > 0)
				continue;
			else if (stackHeights[i] < stackHeights[curr])
				continue;
			return true;
		}
	}
	return false;
}

/// Quickly determine whether one of players has won, ie. run out of one type of stones or has no possibility to capture
ALWAYS_INLINE
int IsEndOfGame(void)
{
	ASSERT2(value == 0, "IsEndOfGame, pos val %d", value);
	FOR(i, 0, 3) if (counts[i] == 0) {
		DPRINT2("black run out of %d", i);
		return WHITE;
	}
	FOR(i, 4, 7) if (counts[i] == 0) {
		DPRINT2("white run out of %d", i);
		return BLACK;
	}
	if (moveNumber == 1) {
		int index = (1 - player) * 2;
		if (zoneOfControl[index] + zoneOfControl[index + 1] + zoneOfControl[index + 2] == 0) {
			ASSERT2(!HasLegalMoves(), "sum of ZOC for %d is 0, but he has legal moves!!!", player);
			DPRINT2("no legal moves for %d", player);
			DBG2(printZOCDebug());
			return -player;

		}
		//else ASSERT2(HasLegalMoves(), "sum of ZOC for %d > 0, but he has NO legal moves!!!", player);
	}
	return false;
}

void printHighestDebug(void)
{		// for debugging highestStack
	FOR(i, 0, STONE_TYPES) {
		DPRINT("type %d, highest %d, pos %d", i, highestStack[i], highestStackPos[i]);
		FOR(j, 0, MAX_STACK_HEIGHT) {
			if (countsByHeight[i][j] > 0)
				printf("%d:%d, ", j, countsByHeight[i][j]);
		}
		printf("\n");
	}
}

/// Executes moves and updates state, ZOC, hash ...
ALWAYS_INLINE
void ExecuteMove(Move * move)
{
	DPRINT2("move to execute: from %s to %s", IndexToFieldName(move->from), IndexToFieldName(move->to));
	ASSERT2(IsMovePossible(move), "EXEC: move not possible!!");
	ASSERT2(value == 0, "EXEC: pos value %d", value);
	ASSERT2(!IsEndOfGame(), "EXEC: end of game");
	int from = move->from;
	int to = move->to;
	move->value = materialValue;
	ASSERT2(from < 0 || from != to, "exec: from %d to %d", from, to);
	if (from == -1) {	//pass
		ASSERT2(moveNumber == 2, "pass is not allowed when moveNumber == 1");
		//do nothing
		DPRINT2("pass ...");
	} else {
		// save old state
		move->oldStone = board[to];
		move->oldStackHeight = stackHeights[to];
		// update hash -- delete old pieces on fields from and to
		hash ^= HashedPositions[to][move->oldStone + 3][move->oldStackHeight];
		hash ^= HashedPositions[from][board[from] + 3][stackHeights[from]];
		// update ZOC
		ZOCRemovingStoneFrom(from, to);
		ZOCRemovingStoneTo(to);
		if (threatenByCounts[from] > 0)
			zoneOfControl[board[from] + 3]--;
		threatenByCounts[from] = 0;
		if (threatenByCounts[to] > 0)
			zoneOfControl[board[to] + 3]--;
		threatenByCounts[to] = 0;
		// updates highest stack
		int hFrom = stackHeights[from];
		int hTo = stackHeights[to];
		int stoneIndexFrom = board[from] + 3;
		int stoneIndexTo = board[to] + 3;
		ASSERT2(stoneIndexFrom != 3, "exec stoneIndexFrom %d", stoneIndexFrom);
		ASSERT2(stoneIndexTo != 3, "exec stoneIndexTo %d", stoneIndexTo);
		countsByHeight[stoneIndexFrom][hFrom]--;
		countsByHeight[stoneIndexTo][hTo]--;
		if (hTo == highestStack[stoneIndexTo]) {
			int hMax = hTo;
			while (countsByHeight[stoneIndexTo][hMax] == 0 && hMax > 0)
				hMax--;
			highestStack[stoneIndexTo] = hMax;
			//if (stoneSum >= VALUEABLE_STACK_COUNT_FROM) { // TODO if not running PNS
			if (highestStackPos[stoneIndexTo] == to && hMax > 0) {
				int cached = highestStackPosCache[stoneIndexTo][hMax];
				if (cached != -1 && cached != from && cached != to && stackHeights[cached] == hMax && board[cached] == stoneIndexTo - 3) {
					highestStackPos[stoneIndexTo] = cached;
				} else {
					bool found = false;
					FOR(i,0,BOARD_ARRAY_SIZE) {
						if (board[i] == stoneIndexTo - 3 && stackHeights[i] == hMax && i != from && i != to) { // TODO on fields from and to there are (maybe) not accurate values
							highestStackPos[stoneIndexTo] = i;
							highestStackPosCache[stoneIndexTo][hMax] = i;
							found = true;
							break;
						}
					}
					ASSERT(found, "exec: HIGHEST STACK NOT FOUND, h %d, type %d", hMax, stoneIndexTo - 3); // TODO turn off this ASSERT
				}
			}
		}
		// capturing
		if (board[from] * board[move->to] < 0) {
			// recalculate value
			materialValue += player * (StackQuality(move->oldStackHeight, move->oldStone, to)
						   - StackValueByField[hFrom][from] * FIELD_VALUE_MULT
						   + StackValueByField[hFrom][to] * FIELD_VALUE_MULT);
			stackHeights[to] = stackHeights[from];
		}
		// stacking
		else {
			// recalculate value
			materialValue += player * (-StackQuality(move->oldStackHeight, move->oldStone, to)
						   - StackQuality(hFrom, board[from], from)
						   + StackQuality(hTo + hFrom, board[from], to));
			stackHeights[to] += stackHeights[from];

		}
		// highest stacks
		int hNew = stackHeights[to];
		countsByHeight[stoneIndexFrom][hNew]++;
		if (hNew >= highestStack[stoneIndexFrom]) { // TODO sharp inequality is maybe sufficient
			highestStack[stoneIndexFrom] = hNew;
			highestStackPos[stoneIndexFrom] = to;
			highestStackPosCache[stoneIndexFrom][hNew] = to;
		}
		counts[board[to] + 3]--;
		board[to] = board[from];
		// clear from field
		board[from] = EMPTY;
		stackHeights[from] = 0;
		ASSERT2(HashedPositions[from][EMPTY + 3][0] != 0
			&& HashedPositions[to][board[to] + 3][stackHeights[to]] != 0, "hash 0");
		// update hash -- add new stacks on fields from and to
		hash ^= HashedPositions[from][EMPTY + 3][0];
		hash ^= HashedPositions[to][board[to] + 3][stackHeights[to]];
		//update ZOC
		ZOCAddedStoneTo(to);
		stoneSum--;
	}
	// this is also done for pass move
	history[turnNumber * 2 + moveNumber] = move;	// save executed move to history
	// update player, move number and moveNumber
	moveNumber++;
	if (moveNumber == 3) { // the 1st turn of white player has also two moves for simplicity
		moveNumber = 1;
		hash ^= 1;
		player = -player;
		turnNumber++;
	}
	// is end of game?
	int winner = 0;
	if ((winner = IsEndOfGame()) != 0) {
		value = WIN * winner;
	} else
		value = 0;	//TODO treat with the variable value in another way?
}

/// Reverts last move in history, updates states, ZOC, hash, material value ...
ALWAYS_INLINE
void RevertLastMove(void)
{
	ASSERT2(turnNumber * 2 + moveNumber - 1 >= 0, "revert: turnNumber * 2 + moveNumber - 1 == %d",
		turnNumber * 2 + moveNumber - 1);
	Move *move = history[turnNumber * 2 + moveNumber - 1];	// retrieve move from history
	ASSERT2(move != NULL, "RevertLastMove: move NULL on %d", turnNumber * 2 + moveNumber - 1);
	DPRINT2("move to revert: from %s to %s; turnNumber * 2 + moveNumber - 1 == %d", IndexToFieldName(move->from),
		IndexToFieldName(move->to), turnNumber * 2 + moveNumber - 1);
	DPRINT2("reverting start");
	if (move->from != -1) {	// move is not pass
		int from = move->from;
		int to = move->to;
		ASSERT2(from != to, "revert: from == to == %d", from);
		// update hash -- delete old stacks from it
		hash ^= HashedPositions[from][EMPTY + 3][0];
		hash ^= HashedPositions[to][board[to] + 3][stackHeights[to]];
		//update ZOC
		ZOCRemovingStoneTo(to);
		ASSERT2(threatenByCounts[to] == 0, "revert: threatenByCounts[to]: %d", threatenByCounts[to]);
		if (threatenByCounts[to] > 0)
			zoneOfControl[board[to] + 3]--;
		threatenByCounts[to] = 0;
		int oldStoneIndex = move->oldStone + 3;
		ASSERT2(oldStoneIndex != 3, "revert oldStoneIndex %d, from %d, to %d", oldStoneIndex, from, to);
		// counts and highest stacks
		counts[oldStoneIndex]++;
		int hOld = stackHeights[to];
		int stoneIndexTo = board[to] + 3;
		ASSERT2(stoneIndexTo != 3, "revert stoneIndexTo %d, from %d, to %d", stoneIndexTo, from, to);
		countsByHeight[stoneIndexTo][hOld]--;
		if (hOld == highestStack[stoneIndexTo]) {
			int hMax = hOld;
			while (countsByHeight[stoneIndexTo][hMax] == 0 && hMax > 0)
				hMax--;
			highestStack[stoneIndexTo] = hMax;
			if (highestStackPos[stoneIndexTo] == to && hMax > 0) {
				int cached = highestStackPosCache[stoneIndexTo][hMax];
				if (cached != -1 && cached != from && cached != to && stackHeights[cached] == hMax && board[cached] == stoneIndexTo - 3) {
					highestStackPos[stoneIndexTo] = cached;
				} else {
					bool found = false;
					FOR(i,0,BOARD_ARRAY_SIZE) {
						if (board[i] == stoneIndexTo - 3 && stackHeights[i] == hMax && i != from && i != to) {
							highestStackPos[stoneIndexTo] = i;
							highestStackPosCache[stoneIndexTo][hMax] = i;
							found = true;
							break;
						}
					}
					ASSERT(found, "revert: HIGHEST STACK NOT FOUND, h %d, type %d", hMax, stoneIndexTo - 3); // TODO turn off this ASSERT
				}
			}
		}
		// revert stack height
		bool capt = move->oldStone * board[to] < 0;
		if (capt) {	// capture
			stackHeights[from] = hOld;
		} else {	// stacking
			stackHeights[from] = hOld - move->oldStackHeight;
		}
		// revert highest stacks
		int hNew = move->oldStackHeight;
		countsByHeight[oldStoneIndex][hNew]++;
		if (hNew >= highestStack[oldStoneIndex]) { // TODO sharp inequality is maybe sufficient
			highestStack[oldStoneIndex] = hNew;
			highestStackPos[oldStoneIndex] = to;
			highestStackPosCache[oldStoneIndex][hNew] = to;
		}
		int hFrom = stackHeights[from];
		countsByHeight[stoneIndexTo][hFrom]++;
		if (hFrom >= highestStack[stoneIndexTo]) { // TODO sharp inequality is maybe sufficient
			highestStack[stoneIndexTo] = hFrom;
			highestStackPos[stoneIndexTo] = from;
			highestStackPosCache[stoneIndexTo][hFrom] = from;
		}
		// update board
		board[from] = board[to];
		board[to] = move->oldStone;
		stackHeights[to] = move->oldStackHeight;

		ASSERT2(HashedPositions[to][move->oldStone + 3][move->oldStackHeight] != 0
			&& HashedPositions[from][board[from] + 3][stackHeights[from]] != 0, "hash 0");
		// update hash -- add new stacks to it
		hash ^= HashedPositions[to][move->oldStone + 3][move->oldStackHeight];
		hash ^= HashedPositions[from][board[from] + 3][stackHeights[from]];
		// update ZOC
		ZOCAddedStoneFrom(from, to);
		ZOCAddedStoneTo(to);
		if (capt) {
			DPRINT2("threatenByCounts[to] %d threatenByCounts[from] %d", threatenByCounts[to],
				threatenByCounts[from]);
			threatenByCounts[to]--;	//zapocteno dvakrat
			if (stackHeights[from] == stackHeights[to])
				threatenByCounts[from]--;	//zapocteno dvakrat
		}
		// revert material value
		materialValue = move->value;
		stoneSum++;
	}
	// revert player
	if (moveNumber == 1) {
		hash ^= 1; // TODO is this useful?
		turnNumber--;
		player = -player;
		moveNumber = 2;
	} else
		moveNumber = 1;
	value = 0;
	ASSERT2(!IsEndOfGame(), "after REVERT: is end of game");
}

/// Return index (in board array) of a field determined by its name, ie. A1 or E6
/// It's slow
int FieldNameToIndex(const char *field)
{
	if (field == NULL)
		return INVALID_INPUT;
	for (int i = 0; i < BOARD_ARRAY_SIZE; i++) {
		if (toupper((int) field[0]) == toupper((int) FieldNames[i][0]) && field[1] == (int) FieldNames[i][1])
			return i;
	}
	return INVALID_INPUT;
}

/// Return name of field on given index in the board array, ie. A1 or E6
const char *IndexToFieldName(int index)
{
	if (index < 0 || index >= BOARD_ARRAY_SIZE)
		return "-";
	return FieldNames[index];
}

/// Generate all moves and return linked list of them (moveNumber could be 1 or 2)
inline void GenerateAllMoves(Move ** moves)
{
	Move *m = NULL;
	for (int i = 0; i < BOARD_ARRAY_SIZE; i++) {
		if (board[i] == BORDER || board[i] == EMPTY || board[i] * player < 0)
			continue;
		for (int j = 0; j < DIRECTION_COUNT; j++) {
			int cx = i % 9 + dxs[j];
			int cy = i / 9 + dys[j];
			int curr = cy * 9 + cx;
			while (curr >= 0 && curr < BOARD_ARRAY_SIZE && board[curr] == EMPTY) {
				cx += dxs[j];
				cy += dys[j];
				curr = cy * 9 + cx;
			}
			if (curr < 0 || curr >= BOARD_ARRAY_SIZE || board[curr] == BORDER)
				continue;
			bool capt = board[i] * board[curr] < 0;
			if (!capt) {
				if (moveNumber == 1)
					continue;
				if (counts[board[curr] + 3] <= 1)
					continue;	//dont stack on the last piece, better is to pass
			} else if (stackHeights[i] < stackHeights[curr])
				continue;
			Move *n = (Move *) malloc(sizeof(Move));
			DBG(moveAlive++);
			n->from = i;
			n->to = curr;
			n->next = m;
			ASSERT2(IsMovePossible(n), "Gen all moves: move not possible");
			m = n;
		}
	}
	if (moveNumber == 2) {
		Move *n = (Move *) malloc(sizeof(Move));
		DBG(moveAlive++);
		n->from = -1;	//n->to =
		n->next = m;
		m = n;
	}
	*moves = m;
}

Move *moveArrayToSort[MAX_POSSIBILITIES]; // array for sorting moves

/// Helping method for qsort from stdlib.h
static int compareMoves(const void *a, const void *b)
{
	return (*(Move **) a)->value - (*(Move **) b)->value;
}

/// Generate all moves, sort them heuristically and return linked list of them (moveNumber could be 1 or 2)
/// Used by DFPNS -- it has a little different move values than GenerateAllMovesSortedMove1/2
inline void GenerateAllMovesSorted(Move ** moves)
{
	ASSERT2(abs(value) < WIN, " -- generating moves in winning position!!, moveNumber %d", moveNumber);
	DPRINT2("gen moves sorted, pl %d", player);
	int count = 0;		//captures = 0,
	for (int i = 0; i < BOARD_ARRAY_SIZE; i++) {
		if (board[i] == BORDER || board[i] == EMPTY || board[i] * player < 0)
			continue;
		//DPRINT2("field %d, pl %d", i, player);
		for (int j = 0; j < DIRECTION_COUNT; j++) {
			int cx = i % 9 + dxs[j];
			int cy = i / 9 + dys[j];
			int curr = cy * 9 + cx;
			while (curr >= 0 && curr < BOARD_ARRAY_SIZE && board[curr] == EMPTY) {
				cx += dxs[j];
				cy += dys[j];
				curr = cy * 9 + cx;
			}
			if (curr < 0 || curr >= BOARD_ARRAY_SIZE || board[curr] == BORDER)
				continue;
			bool capt = board[i] * board[curr] < 0;
			if (!capt) {
				if (moveNumber == 1)
					continue;
				if (counts[board[curr] + 3] <= 1)
					continue;	//dont stack on last piece
			} else if (stackHeights[i] < stackHeights[curr])
				continue;
			Move *n = (Move *) malloc(sizeof(Move));
			DBG(moveAlive++);
			n->from = i;
			n->to = curr;
			if (capt) {
				n->value = SORT_CAPTURE_COUNT_MULT * counts[board[curr] + 3] - CapturingStackHeightAdvantage[stackHeights[curr]];
			} else {
				n->value = SORT_STACK_BONUS + SORT_STACK_COUNT_MULT * counts[board[i] + 3];
			}
			ASSERT2(IsMovePossible(n), "Gen all moves sorted: move not possible");
			moveArrayToSort[count++] = n;
		}
	}
	if (count == 0) {
		Move *n = (Move *) malloc(sizeof(Move));
		DBG(moveAlive++);
		n->from = n->to = -1;
		n->next = NULL;
		*moves = n;
		return;
	}
	qsort(moveArrayToSort, count, sizeof(Move *), compareMoves);
	for (int i = 1; i < count; i++) {
		moveArrayToSort[i - 1]->next = moveArrayToSort[i];
	}
	if (moveNumber == 2) {	//pass move
		Move *n = (Move *) malloc(sizeof(Move));
		DBG(moveAlive++);
		n->from = n->to = -1;
		n->next = NULL;
		moveArrayToSort[count - 1]->next = n;
	} else
		moveArrayToSort[count - 1]->next = NULL;
	*moves = moveArrayToSort[0];
}

/// When the position is in moveNumber 1, generate all moves, sort them by heuristics and return a linked list of them
/// Used by AlphaBeta -- it has a little different move values than GenerateAllMovesSorted
inline void GenerateAllMovesSortedMove1(Move ** moves)
{
	ASSERT2(abs(value) < WIN, " -- generating moves in winning position!!, moveNumber %d", moveNumber);
	DPRINT2("gen moves sorted, pl %d", player);
	ASSERT2(moveNumber == 1, "gen all moves sorted moveNumber NOT 1, but %d", moveNumber);
	int count = 0;
	for (int i = 0; i < BOARD_ARRAY_SIZE; i++) {
		if (board[i] == BORDER || board[i] == EMPTY || board[i] * player < 0)
			continue;
		for (int j = 0; j < DIRECTION_COUNT; j++) {
			int cx = i % 9 + dxs[j];
			int cy = i / 9 + dys[j];
			int curr = cy * 9 + cx;
			while (curr >= 0 && curr < BOARD_ARRAY_SIZE && board[curr] == EMPTY) {
				cx += dxs[j];
				cy += dys[j];
				curr = cy * 9 + cx;
			}
			if (curr < 0 || curr >= BOARD_ARRAY_SIZE || board[curr] == BORDER)
				continue;
			bool capt = board[i] * board[curr] < 0;
			if (!capt) {
				continue;
			} else if (stackHeights[i] < stackHeights[curr])
				continue;
			Move *n = (Move *) malloc(sizeof(Move));
			DBG(moveAlive++);
			n->from = i;
			n->to = curr;
			n->value =
			    SORT_CAPTURE_COUNT_MULT * counts[board[curr] + 3] -
			    CapturingStackHeightAdvantage[stackHeights[curr]]
			    - (SORT_HISTORY_PRUNES_MULT * historyPruneMoves[i][curr]);	// + (historyButterflyMoves[i][curr]); //[depth] //TODO try minus
			ASSERT2(IsMovePossible(n), "Gen all moves sorted: move not possible");
			moveArrayToSort[count++] = n;
		}
	}
	DPRINT2("sorting, pl %d", player);
	ASSERT2(count > 0, "gen all moves sorted moveNumber 1, count %d", count);
	qsort(moveArrayToSort, count, sizeof(Move *), compareMoves);
	for (int i = 1; i < count; i++) {
		moveArrayToSort[i - 1]->next = moveArrayToSort[i];
	}
	moveArrayToSort[count - 1]->next = NULL;
	*moves = moveArrayToSort[0];
}

/// When the position is in moveNumber 2, generate all moves, sort them by heuristics and return a linked list of them
/// Used by AlphaBeta -- it has a little different move values than GenerateAllMovesSorted
inline void GenerateAllMovesSortedMove2(Move ** moves)
{
	ASSERT2(abs(value) < WIN, " -- generating moves in winning position!!, moveNumber %d", moveNumber);
	DPRINT2("gen moves sorted, pl %d", player);
	ASSERT2(moveNumber == 2, "gen all moves sorted moveNumber NOT 2, but %d", moveNumber);
	int count = 0;
	for (int i = 0; i < BOARD_ARRAY_SIZE; i++) {
		if (board[i] == BORDER || board[i] == EMPTY || board[i] * player < 0)
			continue;
		for (int j = 0; j < DIRECTION_COUNT; j++) {
			int cx = i % 9 + dxs[j];
			int cy = i / 9 + dys[j];
			int curr = cy * 9 + cx;
			while (curr >= 0 && curr < BOARD_ARRAY_SIZE && board[curr] == EMPTY) {
				cx += dxs[j];
				cy += dys[j];
				curr = cy * 9 + cx;
			}
			if (curr < 0 || curr >= BOARD_ARRAY_SIZE || board[curr] == BORDER)
				continue;
			bool capt = board[i] * board[curr] < 0;
			if (!capt) {
				if (counts[board[curr] + 3] <= 1)
					continue;	//dont stack on last piece
			} else if (stackHeights[i] < stackHeights[curr])
				continue;
			Move *n = (Move *) malloc(sizeof(Move));
			DBG(moveAlive++);
			n->from = i;
			n->to = curr;
			if (capt) {
				n->value = SORT_CAPTURE_COUNT_MULT * counts[board[curr] + 3] - CapturingStackHeightAdvantage[stackHeights[curr]];
			} else {
				n->value = SORT_STACK_BONUS + SORT_STACK_COUNT_MULT * counts[board[i] + 3];
			}
			ASSERT2(IsMovePossible(n), "Gen all moves sorted: move not possible");
			moveArrayToSort[count++] = n;
		}
	}
	DPRINT2("sorting, pl %d", player);
	if (count == 0) {
		// pass move
		Move *n = (Move *) malloc(sizeof(Move));
		DBG(moveAlive++);
		n->from = n->to = -1;
		n->next = NULL;
		*moves = n;
		return;
	}
	qsort(moveArrayToSort, count, sizeof(Move *), compareMoves);
	for (int i = 1; i < count; i++) {
		moveArrayToSort[i - 1]->next = moveArrayToSort[i];
	}
	//pass move
	Move *n = (Move *) malloc(sizeof(Move));
	DBG(moveAlive++);
	n->from = n->to = -1;
	n->next = NULL;
	moveArrayToSort[count - 1]->next = n;	// count > 0 is ensured
	*moves = moveArrayToSort[0];
}

/// Generate all moves, sort them by heuristics and return only maxMoves best
inline void GenerateBestMovesSorted(Move ** moves, int maxMoves)
{
	int count = 0;
	for (int i = 0; i < BOARD_ARRAY_SIZE; i++) {
		if (board[i] == BORDER || board[i] == EMPTY || board[i] * player < 0)
			continue;
		for (int j = 0; j < DIRECTION_COUNT; j++) {
			int cx = i % 9 + dxs[j];
			int cy = i / 9 + dys[j];
			int curr = cy * 9 + cx;
			while (curr >= 0 && curr < BOARD_ARRAY_SIZE && board[curr] == EMPTY) {
				cx += dxs[j];
				cy += dys[j];
				curr = cy * 9 + cx;
			}
			if (curr < 0 || curr >= BOARD_ARRAY_SIZE || board[curr] == BORDER)
				continue;
			bool capt = board[i] * board[curr] < 0;
			if (!capt) {
				if (moveNumber == 1)
					continue;
				if (counts[board[curr] + 3] <= 1)
					continue;	//dont stack on last piece
			} else if (stackHeights[i] < stackHeights[curr])
				continue;
			Move *n = (Move *) malloc(sizeof(Move));
			DBG(moveAlive++);
			n->from = i;
			n->to = curr;
			if (capt) {
				n->value = 10 * counts[board[curr] + 3] - CapturingStackHeightAdvantage[stackHeights[curr]];
			} else {
				n->value = 19 + 3 * counts[board[i] + 3];
			}
			ASSERT2(IsMovePossible(n), "Gen some moves: move not possible");
			moveArrayToSort[count++] = n;
		}
	}
	if (count == 0 && moveNumber == 1) {
		*moves = NULL;
		return;
	}
	qsort(moveArrayToSort, count, sizeof(Move *), compareMoves);
	int min = MIN(count, maxMoves);
	for (int i = 1; i < min; i++) {
		moveArrayToSort[i - 1]->next = moveArrayToSort[i];
	}
	for (int i = min; i < count; i++) {
		FreeMove(moveArrayToSort[i]);
	}
	if (moveNumber == 2 && count < maxMoves) {	//pass move
		Move *n = (Move *) malloc(sizeof(Move));
		DBG(moveAlive++);
		n->from = n->to = -1;
		n->next = NULL;
		if (min > 0)
			moveArrayToSort[min - 1]->next = n; // put pass move to the end of moves list (it's not good move mostly)
		else {
			*moves = n;
			return;
		}
	} else
		moveArrayToSort[min - 1]->next = NULL;
	*moves = moveArrayToSort[0];
}

/// Allocates memory for new move and copies move in the parameter into its variables
ALWAYS_INLINE
Move *CloneMove(Move * move)
{
	ASSERT2(move != NULL, "move to clone NULL");
	Move *m = MALLOC(Move);
	m->from = move->from;
	m->to = move->to;
	m->next = move->next;
	m->oldStackHeight = move->oldStackHeight;
	m->oldStone = move->oldStone;
	m->value = move->value;
	return m;
}
