/*
 * WALTZ
 *
 * Copyright  2012-2013 Pavel Vesely <vesely@kam.mff.cuni.cz>
 *
 * The module saveload contains functions for loading and saving positions
 * and a function for saving best moves.
 *
 * This program is distributed under the GNU General Public License.
 * See COPYING for details.
 */
#include "ai/saveload.h"

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

/// Save position also with counted information like ZOC, hash, stack counts ...
UNUSED
static int SaveWholePosition(const char *fileName)
{
	ASSERT(fileName != NULL, "save: fileName NULL");
	FILE *f = fopen(fileName, "w");
	if (f == NULL) {
		printf("Cannot save to file '%s'\n", fileName);
		return ERROR;
	}
	fprintf(f, "%d %d %d %d %d %lu\n", turnNumber, player, moveNumber, materialValue, value, hash);
	FOR(i, 0, BOARD_ARRAY_SIZE) {
		fprintf(f, "%d ", board[i]);
	}
	fprintf(f, "\n");
	FOR(i, 0, BOARD_ARRAY_SIZE) {
		fprintf(f, "%d ", threatenByCounts[i]);
	}
	fprintf(f, "\n");
	FOR(i, 0, BOARD_ARRAY_SIZE) {
		fprintf(f, "%d ", stackHeights[i]);
	}
	fprintf(f, "\n");
	FOR(i, 0, STONE_TYPES) {
		fprintf(f, "%d ", counts[i]);
	}
	fprintf(f, "\n");
	FOR(i, 0, STONE_TYPES) {
		fprintf(f, "%d ", zoneOfControl[i]);
	}
	fprintf(f, "\n");
	if (fclose(f) == EOF) {
		DPRINT("error closing file %s", fileName);
		return ERROR;
	}
	return OK;
}

/// Save only basic information about the position
int SavePosition(const char *fileName)
{
	ASSERT(fileName != NULL, "save: fileName NULL");
	FILE *f = fopen(fileName, "w");
	if (f == NULL) {
		printf("Cannot save to file '%s'\n", fileName);
		return ERROR;
	}
	if (abs(value) == WIN) {
		fprintf(f, "END, SOMEONE HAS WON, value:\n%d\n", value);
	} else {
		ASSERT(!IsEndOfGame(), "saving end position!! val %d", value);
	}
	fprintf(f, "%d\n", player); //now moveNumber should be always 1 (it's because of first move and simplicity)
	FOR(i, 0, BOARD_ARRAY_SIZE) {
		if (i > 0 && i % 9 == 0)
			fprintf(f, "\n");
		fprintf(f, "%d ", board[i]);
	}
	fprintf(f, "\n");
	FOR(i, 0, BOARD_ARRAY_SIZE) {
		if (i > 0 && i % 9 == 0)
			fprintf(f, "\n");
		fprintf(f, "%d ", stackHeights[i]);
	}
	fprintf(f, "\n");
	if (fclose(f) == EOF) {
		DPRINT("error closing file %s", fileName);
		return ERROR;
	}
	return OK;
}

/// Save basic information about the position
int LoadPosition(const char *fileName)
{
	ASSERT(fileName != NULL, "open: fileName NULL");
	FILE *f = fopen(fileName, "r");
	if (f == NULL) {
		printf("Cannot open file '%s'\n", fileName);
		return ERROR;
	}
	char c1, c2;
	fscanf(f, "%c%c", &c1, &c2);
	if (c1 == 'E') return -42;
	else if (c1 == '1') player = WHITE;
	else player = BLACK;
	DPRINT2("player %d", player);
	FOR(i, 0, STONE_TYPES) {
		counts[i] = 0;
		highestStack[i] = 0;
		highestStackPos[i] = 0;
		FOR(j, 0, MAX_STACK_HEIGHT) {
			countsByHeight[i][j] = 0;
			highestStackPosCache[i][j] = -1;
		}
	}
	int cnt = 0;
	FOR(i, 0, BOARD_ARRAY_SIZE) {
		fscanf(f, "%d", &(board[i]));
		if (board[i] != 100) {
			counts[board[i] + 3]++;
			cnt++;
		}
	}
	FOR(i, 0, BOARD_ARRAY_SIZE) {
		int c;
		fscanf(f, "%d", &c);
		stackHeights[i] = c;
		if (board[i] != BORDER) {
			countsByHeight[board[i] + 3][c]++;
			if (c > highestStack[board[i] + 3]) {
				highestStack[board[i] + 3] = c;
				highestStackPos[board[i] + 3] = i;
				highestStackPosCache[board[i] + 3][c] = i;
			}
		}
	}
	value = 0;
	stoneSum = 0;
	FOR(i, 0, 7) stoneSum += counts[i];
	stoneSum -= counts[3];
	turnNumber = 1;	//no special handeling for the first move
	moveNumber = 1;
	CountZoneOfControl();
	CountMaterialValue();
	CountHash();
	DBG2(printZOCDebug());
	DBG2(printHighestDebug());
	if (fclose(f) == EOF) {
		DPRINT("error closing file %s", fileName);
		return ERROR;
	}
	ASSERT(!IsEndOfGame(), "LOAD END POSITION!!!");
	return OK;
}

int SaveBestMoves(const char *fileName, Move * m1, Move * m2, bool m1cannotBeNull)
{
	DBG2(printHighestDebug());
	ASSERT(fileName != NULL, "save: fileName NULL");
	ASSERT(!m1cannotBeNull || m1 != NULL, "save: m1 NULL");
	//for the first move, m2 == NULL; and also for winning pos (win with one move)
	FILE *f = fopen(fileName, "w");
	if (f == NULL) {
		printf("Cannot save to file '%s'\n", fileName);
		return ERROR;
	}
	if (m1 == NULL) {
		fprintf(f, "-2\n");
	} else {
		const char *f1f = IndexToFieldName(m1->from);
		const char *f1t = IndexToFieldName(m1->to);
		DPRINT("saving move 1: from %s, to %s", f1f, f1t);
		fprintf(f, "%s %s\n", f1f, f1t);
	}
	if (m2 == NULL) {	//first move of white
		fprintf(f, "-2\n");
	} else if (m2->from == -1) {
		fprintf(f, "-1\n");
	} else {
		const char *f2f = IndexToFieldName(m2->from);
		const char *f2t = IndexToFieldName(m2->to);
		bool capt = board[m2->from] * board[m2->to] < 0;
		// fix -- board is changed after the 1st move
		if (m1 != NULL && (m1->to == m2->from || m1->to == m2->to))
			capt = !capt;
		fprintf(f, "%d %s %s\n", capt, f2f, f2t);
	}
	//saving other information about position and search
	fprintf(f, "%0.3f %d\n", searchDuration, value);
	if (fclose(f) == EOF) {
		DPRINT("error closing file %s", fileName);
		return ERROR;
	}
	return OK;
}

int LoadMoves(const char *fileName, Move **m1, Move **m2) {
	ASSERT(fileName != NULL, "save: fileName NULL");
	FILE *f = fopen(fileName, "r");
	if (f == NULL) {
		printf("Cannot load file '%s'\n", fileName);
		return ERROR;
	}
	char m1f[3], m1t[3];
	fscanf(f, "%c%c %c%c", m1f, m1f + 1, m1t, m1t + 1);
	DPRINT("loading moves:\nmove1 %s %s", m1f, m1t);
	int from = FieldNameToIndex(m1f);
	int to = FieldNameToIndex(m1t);
	*m1 = MALLOC(Move);
	(*m1)->from = from;
	(*m1)->to = to;
	int second;
	fscanf(f, "%d", &second);
	if (second != -2) { // a move
		*m2 = MALLOC(Move);
		if (second == -1) { // pass move
			(*m2)->from = -1;
			(*m2)->to = -1;
		}
		else {
			char m2f[3], m2t[3];
			fscanf(f, " %c%c %c%c", m2f, m2f + 1, m2t, m2t + 1);
			DPRINT("move2: %s %s", m2f, m2t);
			from = FieldNameToIndex(m2f);
			to = FieldNameToIndex(m2t);
			(*m2)->from = from;
			(*m2)->to = to;
		}
		
	}
	else {
		*m2 = NULL;
	}
	return OK;
}
